#include "allinea_metric_plugin_api.h"

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/stat.h>

#define INCLUDE_PER_CPU_COUNTERS
typedef int Errno;
#include "cxiSharedSeg.h"

#define ERROR_INITIALIZATION_FAILED 100

#define DEV_SS0 "/dev/ss0"

/*! File descriptor for /dev/ss0. */ 
int ss0_fd = -1;

/*! The average number of cycles spent in each IOP. */
static double   cyclesPerIOP;

/*! The number of cycles spent in IOPs last sample. */
static uint64_t cyclesSpentInIOLastSample;

/*! The number of IOPs last sample.  */
static uint64_t iopsLastSample;

/*! \a 1 if the initial values of the metrics have not been read yet, else \a 0. */
/*!
 *  The first time \a update is called and this variable is set it will
 *  read the initial values of the metrics and then clear this variable.
 */
static int firstTime = 1;

/*! Time of the last sample. */
/*!
 *  If the time of the current sample is different from the time of the last
 *  then we assume it is a new sample and we need to update our reading of the
 *  counters.
 */
static struct timespec lastSampleTime;

/*! This function is called when the metric plugin is loaded. */
/*!
 *  We do not have to restrict ourselves to async-signal-safe functions because
 *  the initialization function will be called without any locks held.
 * 
 *  \param plugin_id an opaque handle for the plugin.
 *  \param unused unused
 *  \return 0 on success; -1 on failure and set errno
 */
int allinea_plugin_initialize(plugin_id_t plugin_id, void *unused)
{
    (void)unused; /* unused variable */

    ss0_fd = open(DEV_SS0, O_RDONLY);
    if (ss0_fd == -1) {
        int saved_errno = errno;
        if (errno == ENOENT)
            allinea_set_plugin_error_messagef(plugin_id, ERROR_INITIALIZATION_FAILED, "%s: no GPFS filesystem detected", DEV_SS0);
        else
            allinea_set_plugin_error_messagef(plugin_id, ERROR_INITIALIZATION_FAILED, "%s: can't access GPFS shared segment device", DEV_SS0);
        errno = saved_errno;
        return -1;
    }
    firstTime = 1;

    return 0;
}

/*! This function is called when the metric plugin is unloaded. */
/*!
 *  We do not have to restrict ourselves to async-signal-safe functions because
 *  the cleanup function will be called without any locks held.
 * 
 *  \param plugin_id an opaque handle for the plugin.
 *  \param unused unused
 *  \return 0 on success; -1 on failure and set errno
 */
int allinea_plugin_cleanup(plugin_id_t id, void *unused)
{
    (void) id;  // Unused parameter
    (void)unused; /* unused variable */

    if (ss0_fd != -1) {
        close(ss0_fd);
        ss0_fd = -1;
    }
    return 0;
}

/*! Called once per sample to read the metrics from /dev/ss0. */
static int update()
{
    int ret, i;
    uintptr_t args[6];
    PerCpuCounters_t buffer;
    
    if (ss0_fd == -1)
        return 0;

    /* Request the VFS statistics counters from the GPFS kernel module via /dev/ss0. */
    args[0] = cxiCounterTypeVfsStatsGetAll;
    args[1] = sizeof(PerCpuCounters_t);
    args[2] = (uintptr_t) &buffer;
    ret = ioctl(ss0_fd, GetCounters, args);
    if (ret != 0)
        return -1;

    /* Add up all the cycles spent in IO across all VFS calls. */
    uint64_t cyclesSpentInIO = 0LL;
    uint64_t iops = 0LL;
    for (i=0;i<nVFSStatItems;++i) {
        cyclesSpentInIO += buffer.vfsstat_count[i].cycles;
        iops            += buffer.vfsstat_count[i].count;
    }

    if (firstTime) {
        cyclesPerIOP = 0;

        firstTime   = 0;
    } else {
        uint64_t iopsSinceLastSample             = iops           - iopsLastSample;
        uint64_t cyclesSpentInIOSinceLastSample = cyclesSpentInIO - cyclesSpentInIOLastSample;

        cyclesPerIOP = (iopsLastSample == 0.0) ? 0.0 :
                           (double) cyclesSpentInIOSinceLastSample / (double) iopsSinceLastSample;
    }

    cyclesSpentInIOLastSample = cyclesSpentInIO;
    iopsLastSample            = iops;
    
    return 0;
}

/*! Get the current value of the given metric. */
/*!
 *  \param metricId the ID of the metric to get the value for
 *  \param inCurrentSampleTime [in] the time the metric was sampled
 *  \param inValue pointer to where the metric is stored
 *  \param outValue [out] value will be written here.
 * 
 *  If this is a new sample (\a lastSampleTime != \a inCurrentsampleTime)
 *  then \a update is called to read the latest counters from
 *  /dev/ss0.
 *
 *  \note inValue is a pointer. The value is not read from the pointer until
 *  *after* \a update has been called. \a update may change the pointed to value
 *  which means the contents of \a *inValue may have changed after a call to this
 *  function.
 */    
static int getMetricValue(metric_id_t metricId, const struct timespec *inCurrentSampleTime, double *inValue, double *outValue)
{
    (void)metricId; /* unused variable */

    if (ss0_fd == -1)
        return 0;
    
    if (lastSampleTime.tv_sec  != inCurrentSampleTime->tv_sec ||
        lastSampleTime.tv_nsec != inCurrentSampleTime->tv_nsec) {

        lastSampleTime.tv_sec  = inCurrentSampleTime->tv_sec;
        lastSampleTime.tv_nsec = inCurrentSampleTime->tv_nsec;

        int ret = update();
        if (ret != 0)
            return ret;
    }
    
    *outValue = *inValue;
    
    return 0;
}

int allinea_gpfsCyclesPerIOP(metric_id_t metricId, struct timespec *inOutCurrentSampleTime, double *outValue)
{
    return getMetricValue(metricId, inOutCurrentSampleTime, &cyclesPerIOP, outValue);
}
